/** 
 *   FIXME ....
 *
 *   (c) Copyright 2004 Federico Bareilles <fede@iar.unlp.edu.ar>,
 *   Instituto Argentino de Radio Astronomia (IAR).
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *
 *   TODO: ???
 **/

#include <stdio.h>

#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include "e5850.h"
#include "encoders.h"
#include "my_stdio.h"
#include "iar.h"
#include "coordinate.h"
#include "anemometer.h"

int fd_local = -1;
int e5850_nrb = -1;
int encoder_num[] = {RA_CASE, DEC_CASE};

struct encoders_data *d = NULL;

/*****************************************************************************
 * local functions:
 **/

static inline int gray2bin(int bits, int binary)
{
	int gray_code = 0;
	int bit_number;
	int bit = 0;
	
	for (bit_number = bits-1; bit_number >= 0; bit_number--) {
		bit ^= binary >> bit_number & 1;
		if(bit)
			gray_code |= (1 << bit_number);
	}
	return gray_code;
}
	

int get_bits_off_resolution(void)
{
	if(fd_local)
		return ioctl(fd_local, E5850_GETBITRES);
	return -1;
}


static inline int diff_evaluate(int old, int new, int *current_factor)
{
	int diff = old - new;
	
	if( diff <  - 0x2000 ) { /* 2^13 transit from W to E or
				    from S to N */
		*current_factor -= 0x4000;
	}
	if ( diff > 0x2000 ) {
		*current_factor += 0x4000;
	}
	return *current_factor + new;
}

int encoders_pipe(int fd)
{
	int     ra  = encoder_num[RA_CASE],
		dec = encoder_num[DEC_CASE];
		
	while(1){
		encoders_get_next_data(NULL);
/*
		fd_printf(fd, "%5d %5d %3d\n", 
			  d->val[ra], d->val[dec], d->anemometer);
*/
		fd_printf(fd, "%c:%8.4f %8.4f %3d\n", POSITION_PREFIX, 
			  ENC2DEG(d->val[ra]), 
			  ENC2DEC(d->val[dec]),
			  anemometer_conv(d->anemometer));

#if 0
		fd_printf(fd,
			  "%5d (%8.4f) %5d (%8.4f) %3d\n", 
			  d->encoder[ra],  
			  ENC2DEG(d->val[ra]),
			  d->encoder[dec], 
			  -ENC2DEG(d->val[dec]) /* + IAR_LATITUDE */,
			  anemometer_conv(d->anemometer));
#endif
	}

	return 0;
}

/*
 * local functions END.
 *****************************************************************************/

int encoders_open(int ra, int dec)
{
	int i;
	int fd[2];
	int childpid;

	fd_local = open("/dev/e5850",O_RDONLY);
	if(fd_local > 0){
		encoder_num[RA_CASE] = ra;
		encoder_num[DEC_CASE] = dec;
		e5850_nrb = get_bits_off_resolution();
		d = (struct encoders_data *) 
			malloc(sizeof(struct encoders_data));
		for(i=0;i<2;i++){
			d->factor[i] = 0;
			d->old[i] = 0;
		}
		pipe(fd);
		if((childpid = fork()) == -1)
		{
			perror("fork");
			exit(1);
		}
		if(childpid == 0) {
			/* Child process closes up input side of pipe */
			close(fd[0]);
			encoders_pipe(fd[1]);
			//exit(0);
		}
		else {
			/* Parent process closes up output side of pipe */
			close(fd[1]);
			return fd[0];
		}
		
	}
	
	return -1;
}


int encoders_close(void)
{

	/* FIXME: kill the pipe process (man 2 kill)*/
	int ret = close(fd_local);
	fd_local = -1;
	return ret;
}


int encoders_buff_len(void)
{
	int len = -1;
	if(fd_local){
		ioctl(fd_local, E5850_GETBUFFLEN, &len);
		len = len >> 4;
	}
	return 1/*len*/;
}


int encoders_clear_buff(void)
{
	if(fd_local)
		return ioctl(fd_local, E5850_CLEARBUFF);
	else
		return -1;
}


int encoders_is_running(void)
{
	if(fd_local)
		return ioctl(fd_local, E5850_ISRUNNING);
	else
		return -1;
}


int encoders_stop(void)
{
	if(fd_local)
		return ioctl(fd_local, E5850_STOP);
	else
		return -1;
}

	
int encoders_start(void)
{
	if(fd_local)
		return ioctl(fd_local, E5850_START);
	else
		return -1;
}


#if 0
int encoders_get_raw_data(struct encoders_data *d)
{
	int i;
	char buff[E5850_DATA_LEN];
	char *np = buff;
	char *endp;

	if ( read(fd_local, buff, E5850_DATA_LEN) != E5850_DATA_LEN)
		return -1;
	for(i=0;i<2;i++){
		d->encoder[i] = 
			gray2bin(e5850_nrb /* 14 */, 
				 (int) strtol(np, &endp, 16 ));
		np = endp + 1;
	}
	d->encoder[i] = (int) strtol(np, &endp, 16 );

	return 1;
}

#endif

int encoders_get_next_data(struct encoders_data *en)
{
	int i;
	char buff[E5850_DATA_LEN];
	char *np = buff;
	char *endp;
#if 1
	if ( read(fd_local, buff, E5850_DATA_LEN) != E5850_DATA_LEN)
		return -1;
#else
	usleep(240000);
	sprintf(buff,"0001 1ffe 0050 \n");
#endif
	for(i=0;i<2;i++){
		d->encoder[i] = 
			gray2bin(e5850_nrb, 
				 (int) strtol(np, &endp, 16 ));
		d->val[i] = diff_evaluate(d->old[i], d->encoder[i],
					  &d->factor[i]);
		d->old[i] = d->encoder[i];

		np = endp + 1;
	}
	d->anemometer = (int) strtol(np, &endp, 16 );
	if( en != NULL) memcpy(en, d, sizeof(struct encoders_data));

	return 1;
}


int get_encoder(int num)
{
	int res = -1;

	/* FIXME: no tienr acceso a la estructura con los factores de
           correccion de ambiguedad de los encoders */
	if(fd_local){
		res = ioctl(fd_local, E5850_GETNVAL, &num);
		if( res == 0 )
			return gray2bin(e5850_nrb, num);
		
	}
	return res;
}


int encoder_set_cw(int num)
{
	if(fd_local)
		return ioctl(fd_local, E5850_SETCW, &num);
	return -1;
}


int encoder_set_ccw(int num)
{
	if(fd_local)
		return ioctl(fd_local, E5850_SETCCW, &num);
	return -1;
}


int encoder_set_zero(int num)
{
	if(fd_local)
		return ioctl(fd_local, E5850_SETZERO, &num);

	return -1;
}


int sidereal_clock_running(void)
{
	if(fd_local)
		return ioctl(fd_local, E5850_SCLKRUNNING);
	return -1;
}


int put_on_fail_condition(void)
{
	if(fd_local)
		return ioctl(fd_local, E5850_FAILMODE);
	return -1;
}


int put_on_normal_condition(void)
{
	/* FIXME: if on normal conditio ioctl return -1, the same
           value off fail */
	if(fd_local)
		return ioctl(fd_local, E5850_NORMALMODE);
	return -1;
}


int get_anemometer(void)
{
	int res = -1;
	int num = 2;

	if(fd_local){
		res = ioctl(fd_local, E5850_GETNVAL, &num);
		if( res == 0 )
			return num;
		
	}
	return res;
}






